#ifndef _MAIN_CPP
#define _MAIN_CPP

#include "basic.h"
#include "DCCASat2014_help.h"
#include "DCCASat2014.h"
#include "FrwCB2014_help.h"
#include "FrwCB2014.h"

#include <sys/times.h> //these two h files are for linux
#include <unistd.h>

char *inst;
int seed;

int solve_mode; // 1 stands for FrwCB2014; 2 stands for DCCASat2014

void choose_normal_FrwCB2014()
{
	init = init_normal_FrwCB2014;
	flip = flip_normal_FrwCB2014;
	pick_var = pick_var_normal_FrwCB2014;
	max_flips = max_flips_FrwCB2014;
	solve_mode = 1;
}

void choose_huge_FrwCB2014()
{
	init = init_huge_FrwCB2014;
	flip = flip_huge_FrwCB2014;
	pick_var = pick_var_huge_FrwCB2014;
	max_flips = max_flips_FrwCB2014;
	solve_mode = 1;
}

void choose_lm_FrwCB2014()
{
	init = init_lm_FrwCB2014;
	flip = flip_lm_FrwCB2014;
	pick_var = pick_var_lm_FrwCB2014;
	max_flips = max_flips_FrwCB2014;
	solve_mode = 1;
}

void choose_3SAT_DCCASat2014()
{
	set_clause_weighting();
	
	init = init_DCCASat2014;
	flip = flip_3SAT_DCCASat2014;
	pick_var = pick_var_3SAT_DCCASat2014;
	max_flips = max_flips_DCCASat2014;
	solve_mode = 2;
}

void choose_large_DCCASat2014()
{
	set_clause_weighting();
	
	init = init_DCCASat2014;
	flip = flip_large_DCCASat2014;
	pick_var = pick_var_large_DCCASat2014;
	max_flips = max_flips_DCCASat2014;
	
	para_d = 13-max_clause_len;
	if(para_d<6) para_d = 6;
	
	solve_mode = 2;
}

void set_functions()
{
	double r;
	r = (double)num_clauses/num_vars;
	
	if(probtype==SAT3)
	{
		if(r<=4.24)
		{
			//employ FrwCB2014
			if(num_vars>20000)
			{
				choose_huge_FrwCB2014();
				//prob = 0.6*MY_RAND_MAX_INT;
			}
			else
			{
				choose_normal_FrwCB2014();
				//prob = 0.6*MY_RAND_MAX_INT;
			}
		}
		else
		{
			//employ DCCASat2014
			choose_3SAT_DCCASat2014();
		}
	}
	else if(probtype==SAT4)
	{
		if(r<=9.35)
		{
			//employ FrwCB2014
			choose_normal_FrwCB2014();
			//prob = 0.65*MY_RAND_MAX_INT;
		}
		else
		{
			//employ DCCASat2014
			choose_large_DCCASat2014();
		}
	}
	else if(probtype==SAT5)
	{
		if(r<=20.1)
		{
			//employ FrwCB2014
			choose_lm_FrwCB2014();
			//prob = 0.58*MY_RAND_MAX_INT;
		}
		else
		{
			//employ DCCASat2014
			choose_large_DCCASat2014();
		}
	}
	else if(probtype==SAT6)
	{
		if(r<=41.2)
		{
			//employ FrwCB2014
			choose_lm_FrwCB2014();
			//prob = 0.69*MY_RAND_MAX_INT;
		}
		else
		{
			//employ DCCASat2014
			choose_large_DCCASat2014();
		}
	}
	else if(probtype==SAT7)
	{
		if(r<=80)
		{
			//employ FrwCB2014
			choose_lm_FrwCB2014();
			//prob = 0.76*MY_RAND_MAX_INT;
		}
		else
		{
			//employ DCCASat2014
			choose_large_DCCASat2014();
		}
	}
	else
	{
		//employ DCCASat2014
		choose_3SAT_DCCASat2014();
	}
}

void local_search()
{
	int flipvar;
	for (step = 0; step<max_flips; step++)
	{
		if(unsat_stack_fill_pointer==0) return;
		
		flipvar = pick_var();
		flip(flipvar);
		time_stamp[flipvar] = step;
	}
}

int parse_parameters(int argc, char* argv[])
{
	int i;
	string cur_str;
	
	if(argc==1)
	{
		return 0;
	}
	
	for(i=1; i<argc; i++)
	{
		cur_str = argv[i];
		if(cur_str.compare("-inst")==0)
		{
			i++;
			inst = argv[i];
		}
		else if(cur_str.compare("-seed")==0)
		{
			i++;
			sscanf(argv[i], "%d", &seed);
		}
		else if(cur_str.compare("-prob")==0)
		{
			i++;
			double p;
			sscanf(argv[i], "%lf", &p);
			prob = p*MY_RAND_MAX_INT;
		}
		else if(cur_str.compare("-w_t")==0)
		{
			i++;
			sscanf(argv[i], "%d", &threshold);
		}
		else if(cur_str.compare("-sp")==0)
		{
			i++;
			sscanf(argv[i], "%f", &smooth_probability);
		}
		else
		{
			return 0;
		}
	}
	return 1;
}

void print_information()
{
	if(solve_mode==1)
	{
		cout<<"c Using FrwCB2014"<<endl;
	}
	else
	{
		cout<<"c Using DCCASat2014"<<endl;
	}
	cout<<"c instance = "<<inst<<endl;
	cout<<"c seed = "<<seed<<endl;
	cout<<"c num_vars = "<<num_vars<<endl;
	cout<<"c num_clauses = "<<num_clauses<<endl;
	cout<<"c max_clause_len = "<<max_clause_len<<endl;
	cout<<"c min_clause_len = "<<min_clause_len<<endl;
	cout<<"c ratio = "<<(double)num_clauses/num_vars<<endl;
	cout<<"c max_tries = "<<max_tries<<endl;
	cout<<"c max_flips = "<<max_flips<<endl;
	
	cout<<"c prob = "<<(double)prob/MY_RAND_MAX_INT<<endl;
	cout<<"c w_t = "<<threshold<<endl;
	cout<<"c sp = "<<smooth_probability<<endl;
}

int main(int argc, char* argv[])
{
	long long i;
	struct tms start, stop;
	times(&start);
	
	cout<<"c this is CSCCSat2014 [Version: CSSC2014]" <<endl;
	
	prob = 600*10000;
	threshold = 300;
	smooth_probability = 0.8f;
	int ret = parse_parameters(argc, argv);
	if(ret==0)
	{
		cout<<"c Invalid parameters!"<<endl;
		cout<<"c Usage: "<<argv[0] <<" -inst <instance> -seed <seed> -prob <prob> -w_t <w_t> -sp <sp>"<<endl;
		return -1;
	}

	if (build_instance(inst)==0)
	{
		cout<<"c Invalid filename: "<< inst <<endl;
		return -1;
	}
	
	if(num_clauses==0)
	{
		cout<<"s SATISFIABLE"<<endl;
		print_solution();
		free_memory();
		return 0;
	}
	
	if(num_empty_clauses>0)
	{
		cout<<"s UNSATISFIABLE"<<endl;
		free_memory();
		return 0;
	}
    
    srand(seed);
    
	set_functions();
	
	print_information();
	
	cout<<"c start searching"<<endl;
	
	for (i = 0; i <= max_tries; i++) 
	{
		 init();
		 
		 times(&stop);
	 
		 local_search();

		 if (unsat_stack_fill_pointer==0) break;
	}

	times(&stop);
	double comp_time = double(stop.tms_utime - start.tms_utime +stop.tms_stime - start.tms_stime) / sysconf(_SC_CLK_TCK);

	if(unsat_stack_fill_pointer==0)
	{
		if(verify_sol()==1)
		{
			cout<<"s SATISFIABLE"<<endl;
			print_solution();
			
			cout<<"c solveSteps = "<<i<<" tries + "<<step<<" steps (each try has " << max_flips << " steps)."<<endl;
            cout<<"c solveTime = "<<comp_time<<endl;
        }
        else 
        {
        	cout<<"s UNKNOWN"<<endl;
        	cout<<"c Sorry, something is wrong."<<endl;
        }
    }
    else
    {
    	cout<<"s UNKNOWN"<<endl;
    }
	 
    free_memory();

    return 0;
}

#endif

